import importlib
import logging
import os

from flask import Flask
from furl import furl
from raven.contrib.flask import Sentry
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool
from flask_mail import Mail


logger = logging.getLogger('factory')
logger.setLevel(logging.INFO)


# !NB: use it till flask-sqlalchemy will be able to use pools
class SQLAlchemyNullPool(SQLAlchemy):
    def apply_driver_hacks(self, app, info, options):
        options.update({
            'poolclass': NullPool,
        })
        super(SQLAlchemyNullPool, self).apply_driver_hacks(app, info, options)


db = SQLAlchemyNullPool()


def init_db(app):
    sqllogger = logging.getLogger('sqlalchemy.engine')
    sqllogger.setLevel(logging.CRITICAL)
    db.init_app(app)
    return db


def init_mail(app):
    mail = Mail()
    mail.init_app(app)
    app.mail = mail


def init_config(app, config=None):
    if config:
        app.config.from_mapping(config)
    else:
        app.config.from_object(get_config_name())


def get_config_name():
    return 'config.' + os.getenv('FLASK_CONFIG', 'default')


def init_sentry(app):
    if not app.debug:
        dsn = app.config.get('SENTRY_DSN')
        Sentry(app, dsn=dsn)


def init_migrations(app):
    return Migrate(app, db)


def register_blueprints(app, blueprints=None):
    try:
        blueprints = blueprints or os.listdir('app')
    except FileNotFoundError:
        logger.error('No app folder found, blueprints were not imported')
        return
    for name in blueprints:
        is_dir = os.path.isdir(os.path.join('app', name))
        is_underscored = name.startswith('_')
        if is_dir and not is_underscored:
            try:
                views = importlib.import_module('app.%s.views' % name)
            except ImportError:
                logger.info('Package %s has no views to import' % name)
                continue
            app.register_blueprint(views.mod)


def create_app(config=None, name=None):
    app = Flask(
        name or __name__.split('.')[0],
        template_folder='../templates',
        static_folder='../static'
    )
    init_config(app, config=config)
    app.db = init_db(app)
    register_blueprints(app)

    init_sentry(app)
    init_migrations(app)
    return app


def create_db(dsn):
    engine, db_name = parse_dsn(dsn)
    logger.info('Creating DB:%s...' % db_name)
    conn = engine.connect()
    conn.execute('CREATE DATABASE %s;' % db_name)
    conn.close()


def create_tables(app):
    conn = app.db.session.connection()
    table_names = app.db.engine.dialect.get_table_names(connection=conn)
    if not table_names:
        logger.info('Creating tables...')
        app.db.create_all()
    else:
        logger.info('Tables are already created %s ' % table_names)


def drop_db(dsn):
    if is_db_exists(dsn):
        engine, db_name = parse_dsn(dsn)
        logger.info('Dropping DB:%s...' % db_name)
        conn = engine.connect()
        conn.execute('DROP DATABASE %s;' % db_name)
        conn.close()


def parse_dsn(dsn):
    f = furl(dsn)
    msg = 'No database name provided in DSN:%s' % dsn
    assert f.path.segments, msg
    assert f.path.segments[0], msg
    db_name = f.path.segments[0]
    # db_user = f.username
    f.path.remove(db_name)  # Remove database name from DSN
    engine = create_engine(f.url, isolation_level='AUTOCOMMIT')
    return engine, db_name


def is_db_exists(dsn):
    engine, name = parse_dsn(dsn)
    query = 'SELECT datname FROM pg_database WHERE datistemplate = false;'
    conn = engine.connect()
    names = {row[0] for row in conn.execute(query).fetchall()}
    conn.close()
    return name in names

